// gftcl.c
// Interfaces between the GoFlight SDK and the Tcl scripting language.
// Full API doc at http://www.hoppie.nl/gfsb747/gftcldll.html

#define dll_version "1.5"

#include <windows.h>
#include <math.h>
#include "..\include\tcl.h"
#include "..\include\gfdevapi.h"


///// Internal Persistant Variables //////////////////////////////////////
Tcl_Interp *INTERP = NULL;  // Tcl interpreter to send the callback to.
Tcl_DString EVENTPROC;      // Tcl script to evaluate at callback time.
// Sane defaults for the lever calibration, which deliver raw data.
int leverMin[6]   = {0,0,0,0,0,0};
int leverMax[6]   = {255,255,255,255,255,255};
int leverJit[6]   = {1,1,1,1,1,1};
int leverSteps[6] = {255,255,255,255,255,255};


///// Internal functions /////////////////////////////////////////////////

void done(ClientData clientData) {
  // Clean up all used resources.
  Tcl_DStringFree(&EVENTPROC);

  // Terminate the GoFlight library.
  GFDev_Terminate();
} // done


BOOL gfEvent(GFDEV_HW_TYPE hwType,
             unsigned short devNr,
             unsigned long inputReport) {
  // Called when a GoFlight input event occurs and a Tcl callback procedure
  // must be evaluated. Since this may take a considerable time, and since
  // the GoFlight library does not protect against re-entrant calling of
  // this function, a poor man's semaphore is used to alleviate most problems.
  // The Tcl procedure is called with {devType devID control value} args.

  Tcl_DString tclCmd;
  char* devType;       // Pointer to constant string in code segment.
  char  input[64]="";  // Will be filled with hardware-dependent data.
  char  params[128];   // To build the complete parameter string in.
  static short busy=0; // Semaphore to record entrance and exit of this
                       // function. Not officially safe.

  if (busy==1) {
    return TRUE;
  }
  busy = 1;

  // Convert the GoFlight HWTYPE to a string.
  // Also convert the inputReport to a string, different per hardware type.
  // C trick: use blocks to clearly separate the local-local variables.
  switch(hwType) {
  case GFDEVHWTYPE_GF166: {
      // Inner dial, outer dial, and three momentary action pushbuttons.
      LPGF166INPUTREPORT gf166input = (LPGF166INPUTREPORT) inputReport;
      devType = "GF166";
      if (gf166input->nLgDialVal!=0) {
        sprintf(input,"outer %d",gf166input->nLgDialVal);
      }
      if (gf166input->nSmDialVal!=0) {
        sprintf(input,"inner %d",gf166input->nSmDialVal);
      }
      if (gf166input->bSwitchState & GF166_BTN_FUNC_L) {
        sprintf(input,"left 1");
      }
      if (gf166input->bSwitchState & GF166_BTN_FUNC_C) {
        sprintf(input,"center 1");
      }
      if (gf166input->bSwitchState & GF166_BTN_FUNC_R) {
        sprintf(input,"right 1");
      }
      break;
    }

    case GFDEVHWTYPE_GF45: {
      // Left and right dial.
      LPGF45INPUTREPORT gf45input = (LPGF45INPUTREPORT) inputReport;
      devType = "GF45";
      if (gf45input->nLDialVal!=0) {
        sprintf(input,"left %d",gf45input->nLDialVal);
      }
      if (gf45input->nRDialVal!=0) {
        sprintf(input,"right %d",gf45input->nRDialVal);
      }
      break;
    }

    case GFDEVHWTYPE_GF46: {
      // Outer and inner dial plus single pushbutton.
      LPGF46INPUTREPORT gf46input = (LPGF46INPUTREPORT) inputReport;
      devType = "GF46";
      if (gf46input->nLgDialVal!=0) {
        sprintf(input,"outer %d",gf46input->nLgDialVal);
      }
      if (gf46input->nSmDialVal!=0) {
        sprintf(input,"inner %d",gf46input->nSmDialVal);
      }
      if (gf46input->bSwitchState & GF46_BTN_FUNC) {
        sprintf(input,"button 1");
      }
      break;
    }

    case GFDEVHWTYPE_GFLGT:
      // Not yet supported.
      devType = "GFLGT";
      break;

    case GFDEVHWTYPE_GFP8: {
      // Button 1 to 8, 1=leftmost. Only report push, not release.
      // Only the rightmost button currently pushed is reported.
      BYTE buttons = *(BYTE*)inputReport;
      int b;
      devType = "GFP8";
      for (b=0; b<8; b++) {
        if (buttons & (1<<b)) {
          sprintf(input,"button%d 1",b+1);
        }
      }
      break;
    }

    case GFDEVHWTYPE_GFT8: {
      // Switch 1 to 8, 1=leftmost. Reports any change from the previous
      // state of the switches (which is kept in local memory).
      // Silly effects when two switches are operated at the same time.
      BYTE switches = *(BYTE*)inputReport;
      static BYTE previous = 0;
      BYTE changed = 0;
      int s;
      devType = "GFT8";
      // Calculate the changes from last known switch positions.
      changed = switches^previous;
      for (s=0; s<8; s++) {
        if (changed & (1<<s)) {
          // This switch has been changed, report its current position.
          sprintf(input,"switch%d %d",s+1,(switches & (1<<s))>>s);
        }
      }
      previous = switches;
      break;
    }

    case GFDEVHWTYPE_GFTQ6: {
      // Six levers A..F and four lever switches B..E (A and F not installed,
      // but the software assumes they could be installed in the future).
      // The levers run (after calibration) from 0 (down) to a given number
      // of steps, typically 100 (up).
      // The switches do 0 or 1, 1 = contact made between the two levers.

      static int prevA,prevB,prevC,prevD,prevE,prevF,prevSwitches = 0;
      int levA,levB,levC,levD,levE,levF,switches,swA,swB,swC,swD,swE,swF;
      LPGFTQ6INPUTREPORT gftq6input = (LPGFTQ6INPUTREPORT) inputReport;
      float f;
      static DWORD lastTime = 0;
      DWORD currentTime;
      devType = "GFTQ6";

      currentTime = GetTickCount();
      if ((currentTime-lastTime)<100) {
        // We do not want more than ten events per second.
        break;
      }
      lastTime = currentTime;

      // Fetch all current lever positions.
      levA = gftq6input->bLeverAVal;
      levB = gftq6input->bLeverBVal;
      levC = gftq6input->bLeverCVal;
      levD = gftq6input->bLeverDVal;
      levE = gftq6input->bLeverEVal;
      levF = gftq6input->bLeverFVal;

      // Fetch the current switch positions.
      switches = gftq6input->bSwitchState;

      // Jitter suppression: any lever must have exceeded its jitter
      // threshold since last reported position change, or any switch
      // must have changed position.
      if ( (abs(levA-prevA)>=leverJit[0]) ||
           (abs(levB-prevB)>=leverJit[1]) ||
           (abs(levC-prevC)>=leverJit[2]) ||
           (abs(levD-prevD)>=leverJit[3]) ||
           (abs(levE-prevE)>=leverJit[4]) ||
           (abs(levF-prevF)>=leverJit[5]) ||
           (switches!=prevSwitches) ) {
        // Significant raw status change detected. Save the new status.
        prevA = levA;
        prevB = levB;
        prevC = levC;
        prevD = levD;
        prevE = levE;
        prevF = levF;
        prevSwitches = switches;

        // Calibrate all levers to a value between 0 and steps. Of course,
        // if the calibration hasn't been done properly, silly values may
        // appear, but nothing should ever break.
        f = (float)leverSteps[0]/(float)(leverMax[0]-leverMin[0]);
        levA = (int)floor((levA-leverMin[0])*f);
        f = (float)leverSteps[1]/(float)(leverMax[1]-leverMin[1]);
        levB = (int)floor((levB-leverMin[1])*f);
        f = (float)leverSteps[2]/(float)(leverMax[2]-leverMin[2]);
        levC = (int)floor((levC-leverMin[2])*f);
        f = (float)leverSteps[3]/(float)(leverMax[3]-leverMin[3]);
        levD = (int)floor((levD-leverMin[3])*f);
        f = (float)leverSteps[4]/(float)(leverMax[4]-leverMin[4]);
        levE = (int)floor((levE-leverMin[4])*f);
        f = (float)leverSteps[5]/(float)(leverMax[5]-leverMin[5]);
        levF = (int)floor((levF-leverMin[5])*f);

        // Extract all individual switches.
        swA = (switches & GFTQ6_SW_A)!=0;
        swB = (switches & GFTQ6_SW_B)!=0;
        swC = (switches & GFTQ6_SW_C)!=0;
        swD = (switches & GFTQ6_SW_D)!=0;
        swE = (switches & GFTQ6_SW_E)!=0;
        swF = (switches & GFTQ6_SW_F)!=0;

        // Produce one single output report.
        sprintf(input,"levers {%3d %3d %3d %3d %3d %3d %d %d %d %d %d %d}",
                      levA,levB,levC,levD,levE,levF,
                      swA,swB,swC,swD,swE,swF);
      } // if jitter threshold exceeded

      break;
    }

    default:
      devType = "UNKNOWN";
  }

  // Generic event handling for all devices.
  if (strcmp(input,"")!=0) {
    // There is known input available.
    // Construct the full Tcl callback command, with parameters.
    sprintf(params," %s %d %s",devType,devNr,input);
    Tcl_DStringInit(&tclCmd);
    Tcl_DStringAppend(&tclCmd,Tcl_DStringValue(&EVENTPROC),-1);
    Tcl_DStringAppend(&tclCmd,params,-1);
    if (Tcl_Eval(INTERP,Tcl_DStringValue(&tclCmd))==TCL_ERROR) {
      // An error occurred. Call the "bgerror" command, if there is one.
      // Otherwise this is a silent failure.
      Tcl_Obj* result;
      result = Tcl_GetObjResult(INTERP);
      sprintf(params,"bgerror {%s}",Tcl_GetString(result));
      Tcl_Eval(INTERP,params);
    }
    Tcl_DStringFree(&tclCmd);
  }

  busy = 0;
  return TRUE;
} // gfEvent


void setGFP8leds(int devNr,int LED,char* val) {
  // Switches one LED on or off.
  // <in> devNr = The GFP8 device number.
  // <in> LED = The LED number, 0=leftmost, 7=rightmost.
  // <in> val = "0"=off, rest is on.
  unsigned char leds;
  unsigned char mask;

  mask = 1<<LED;
  GFP8_GetIndicators(devNr,&leds);
  if (strcmp(val,"0")==0) {
    // LED off.
    leds = leds & ~mask;
  } else {
    // LED on.
    leds = leds | mask;
  }
  GFP8_SetIndicators(devNr,leds);
} // setGFP8leds


void setGFT8leds(int devNr,int LED,char* val) {
  // Switches one LED on or off.
  // <in> devNr = The GFT8 device number.
  // <in> LED = The LED number, 0=leftmost, 7=rightmost.
  // <in> val = "0"=off, rest is on.
  unsigned char leds;
  unsigned char mask;

  mask = 1<<LED;
  GFT8_GetIndicators(devNr,&leds);
  if (strcmp(val,"0")==0) {
    // LED off.
    leds = leds & ~mask;
  } else {
    // LED on.
    leds = leds | mask;
  }
  GFT8_SetIndicators(devNr,leds);
} // setGFT8leds


///// Tcl commands ///////////////////////////////////////////////////////

int versionCmd(ClientData clientData,
               Tcl_Interp *interp,
               int objc,
               Tcl_Obj *CONST objv[]) {
  // Returns the current version of the DLL.
  // <in> Nothing.
  // <out> Version number, such as "1.2".

  char tmp[16];
  Tcl_DString version;

  Tcl_DStringInit(&version);
  sprintf(tmp,"%s",dll_version);
  Tcl_DStringAppend(&version,tmp,-1);

  Tcl_DStringResult(interp,&version);
	return TCL_OK;
} // versionCmd


int devicesCmd(ClientData clientData,
               Tcl_Interp *interp,
               int objc,
               Tcl_Obj *CONST objv[]) {
  // Returns a list containing the numbers of detected GoFlight devices,
  // in a format that is suitable for [array set].
  // All known devices are enumerated, so zeros may occur.
  // At the moment, the device list is NOT hot plug updated!
  // <in> Nothing.
  // <out> List, such as "GF166 3 GF45 1 GFLGT 0 GFP8 0 GFT8 0".

  char tmp[16];
  Tcl_DString devList;

  /*
  // Re-initialise the GoFlight library to have it re-poll all devices.
  // This disables all callbacks!
  if (GFDev_Init(NULL)) {
    Tcl_SetStringObj(Tcl_GetObjResult(interp),
      "Could not re-initialise GoFlight library",-1);
    return TCL_ERROR;
	} */

  // Build the device list.
  Tcl_DStringInit(&devList);

  sprintf(tmp,"GF166 %d",GF166_GetNumDevices());
  Tcl_DStringAppend(&devList,tmp,-1);

  sprintf(tmp," GF45 %d",GF45_GetNumDevices());
  Tcl_DStringAppend(&devList,tmp,-1);

  sprintf(tmp," GF46 %d",GF46_GetNumDevices());
  Tcl_DStringAppend(&devList,tmp,-1);

  sprintf(tmp," GFLGT %d",GFLGT_GetNumDevices());
  Tcl_DStringAppend(&devList,tmp,-1);

  sprintf(tmp," GFP8 %d",GFP8_GetNumDevices());
  Tcl_DStringAppend(&devList,tmp,-1);

  sprintf(tmp," GFT8 %d",GFT8_GetNumDevices());
  Tcl_DStringAppend(&devList,tmp,-1);

  sprintf(tmp," GFTQ6 %d",GFTQ6_GetNumDevices());
  Tcl_DStringAppend(&devList,tmp,-1);

  Tcl_DStringResult(interp,&devList);
	return TCL_OK;
} // devicesCmd


int notifyCmd(ClientData clientData,
            Tcl_Interp *interp,
            int objc,
            Tcl_Obj *CONST objv[]) {
  // Implements the notify command.
  // <in> eventProc = Tcl procedure that is called when a GoFlight
  //                  input event occurs. De-registers when {}.
  //                  The procedure receives the following args:
  //                  {devType devID control value}.
  // <out> Nothing.

  Tcl_DString eventProc;

  // Check for exactly one parameter.
  if (objc!=2) {
    Tcl_WrongNumArgs(interp,1,objv,"eventProc");
    return TCL_ERROR;
  }

  // Get the eventProc. This is a complete Tcl script.
  Tcl_DStringInit(&eventProc);
  Tcl_DStringAppend(&eventProc,Tcl_GetString(objv[1]),-1);

  // If the eventProc=={}, deregister the GoFlight callback.
  if (strcmp(Tcl_DStringValue(&eventProc),"")==0) {
    // Deregister the callback.
    GFDev_UnregisterInputCallback();
    Tcl_DStringFree(&EVENTPROC);
    Tcl_DStringInit(&EVENTPROC);
  } else {
    // Register the callback.
    GFDev_RegisterInputCallback(&gfEvent);
    Tcl_DStringFree(&EVENTPROC);
    Tcl_DStringAppend(&EVENTPROC,Tcl_DStringValue(&eventProc),-1);
  }

  Tcl_DStringFree(&eventProc);
  return TCL_OK;
} // notifyCmd


int gf166Cmd(ClientData clientData,
            Tcl_Interp *interp,
            int objc,
            Tcl_Obj *CONST objv[]) {
  // Implements all GF166 (sub)commands.
  // <in> devNr = GF166 device number, starting with 0.
  // <in> cmd = GF166 subcommand: setLeftDisplay, setRightDisplay,
  //            setLeftLED, setCenterLED, setRightLED, setBrightness.
  // <in> val = A string. For the LEDs, 0 or 1.
  // <out> Nothing.

  int devNr;
  char tmp[256];
  Tcl_DString cmd;
  Tcl_DString val;
  int brightness;

  // Check for exactly three parameters.
  if (objc!=4) {
    Tcl_WrongNumArgs(interp,1,objv,"devNr cmd val");
    return TCL_ERROR;
  }

  // Get the device number and check for sanity.
  if (Tcl_GetIntFromObj(interp,objv[1], &devNr) != TCL_OK) {
    return TCL_ERROR;
  }
  if (devNr<0) {
    // No negative device numbers.
    sprintf(tmp,"Invalid device number %d",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }
  if (devNr>(GF166_GetNumDevices()-1)) {
    // No device numbers higher than the available devices.
    sprintf(tmp,"Device %d unavailable",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the (sub)command.
  Tcl_DStringInit(&cmd);
  Tcl_DStringAppend(&cmd,Tcl_GetString(objv[2]),-1);

  // Get the value.
  Tcl_DStringInit(&val);
  Tcl_DStringAppend(&val,Tcl_GetString(objv[3]),-1);

  // Command ladder.
  if (strcmp(Tcl_DStringValue(&cmd),"setLeftDisplay")==0) {
    // setLeftDisplay value
    GF166_SetLDisplayText(devNr,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setRightDisplay")==0) {
    // setRightDisplay value
    GF166_SetRDisplayText(devNr,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLeftLED")==0) {
    // setLeftLED value
    unsigned char leds;
    GF166_GetIndicators(devNr,&leds);
    if (strcmp(Tcl_DStringValue(&val),"0")==0) {
      // Left LED off.
      leds = leds & ~GF166_LED_FUNC_L;
    } else {
      // Left LED on.
      leds = leds | GF166_LED_FUNC_L;
    }
    GF166_SetIndicators(devNr,leds);

  } else if (strcmp(Tcl_DStringValue(&cmd),"setCenterLED")==0) {
    // setCenterLED value
    unsigned char leds;
    GF166_GetIndicators(devNr,&leds);
    if (strcmp(Tcl_DStringValue(&val),"0")==0) {
      // Center LED off.
      leds = leds & ~GF166_LED_FUNC_C;
    } else {
      // Center LED on.
      leds = leds | GF166_LED_FUNC_C;
    }
    GF166_SetIndicators(devNr,leds);

  } else if (strcmp(Tcl_DStringValue(&cmd),"setRightLED")==0) {
    // setRightLED value
    unsigned char leds;
    GF166_GetIndicators(devNr,&leds);
    if (strcmp(Tcl_DStringValue(&val),"0")==0) {
      // Right LED off.
      leds = leds & ~GF166_LED_FUNC_R;
    } else {
      // Right LED on.
      leds = leds | GF166_LED_FUNC_R;
    }
    GF166_SetIndicators(devNr,leds);

  } else if (strcmp(Tcl_DStringValue(&cmd),"setBrightness")==0) {
    // setBrightness 0..15
    if (Tcl_GetIntFromObj(interp,objv[3], &brightness) != TCL_OK) {
      Tcl_DStringFree(&cmd);
      Tcl_DStringFree(&val);
      return TCL_ERROR;
    }
    GF166_SetBrightness(devNr,brightness);

  } else {
    // Unknown subcommand.
    sprintf(tmp,"Unknown subcommand %s",Tcl_DStringValue(&cmd));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    Tcl_DStringFree(&cmd);
    Tcl_DStringFree(&val);
    return TCL_ERROR;
  }

  Tcl_DStringFree(&cmd);
  Tcl_DStringFree(&val);
  return TCL_OK;
} // gf166Cmd


int gf45Cmd(ClientData clientData,
            Tcl_Interp *interp,
            int objc,
            Tcl_Obj *CONST objv[]) {
  // Implements all GF45 (sub)commands.
  // <in> devNr = GF45 device number, starting with 0.
  // <in> cmd = GF45 subcommand: setAlpha, setNumeric,
  //            setBrightness.
  // <in> val = A string.
  // <out> Nothing.

  int devNr;
  char tmp[256];
  Tcl_DString cmd;
  Tcl_DString val;
  int brightness;

  // Check for exactly three parameters.
  if (objc!=4) {
    Tcl_WrongNumArgs(interp,1,objv,"devNr cmd val");
    return TCL_ERROR;
  }

  // Get the device number and check for sanity.
  if (Tcl_GetIntFromObj(interp,objv[1], &devNr) != TCL_OK) {
    return TCL_ERROR;
  }
  if (devNr<0) {
    // No negative device numbers.
    sprintf(tmp,"Invalid device number %d",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }
  if (devNr>(GF45_GetNumDevices()-1)) {
    // No device numbers higher than the available devices.
    sprintf(tmp,"Device %d unavailable",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the (sub)command.
  Tcl_DStringInit(&cmd);
  Tcl_DStringAppend(&cmd,Tcl_GetString(objv[2]),-1);

  // Get the value.
  Tcl_DStringInit(&val);
  Tcl_DStringAppend(&val,Tcl_GetString(objv[3]),-1);

  // Command ladder.
  if (strcmp(Tcl_DStringValue(&cmd),"setAlpha")==0) {
    // setAlpha value
    GF45_SetAlpha(devNr,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setNumeric")==0) {
    // setNumeric value
    GF45_SetNumeric(devNr,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setBrightness")==0) {
    // setBrightness 0..15
    if (Tcl_GetIntFromObj(interp,objv[3], &brightness) != TCL_OK) {
      Tcl_DStringFree(&cmd);
      Tcl_DStringFree(&val);
      return TCL_ERROR;
    }
    GF45_SetBrightness(devNr,brightness);

  } else {
    // Unknown subcommand.
    sprintf(tmp,"Unknown subcommand %s",Tcl_DStringValue(&cmd));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    Tcl_DStringFree(&cmd);
    Tcl_DStringFree(&val);
    return TCL_ERROR;
  }

  Tcl_DStringFree(&cmd);
  Tcl_DStringFree(&val);
  return TCL_OK;
} // gf45Cmd


int gf46Cmd(ClientData clientData,
            Tcl_Interp *interp,
            int objc,
            Tcl_Obj *CONST objv[]) {
  // Implements all GF46 (sub)commands.
  // <in> devNr = GF46 device number, starting with 0.
  // <in> cmd = GF46 subcommand: setAlpha, setNumeric,
  //            setBrightness.
  // <in> val = A string.
  // <out> Nothing.

  int devNr;
  char tmp[256];
  Tcl_DString cmd;
  Tcl_DString val;
  int brightness;

  // Check for exactly three parameters.
  if (objc!=4) {
    Tcl_WrongNumArgs(interp,1,objv,"devNr cmd val");
    return TCL_ERROR;
  }

  // Get the device number and check for sanity.
  if (Tcl_GetIntFromObj(interp,objv[1], &devNr) != TCL_OK) {
    return TCL_ERROR;
  }
  if (devNr<0) {
    // No negative device numbers.
    sprintf(tmp,"Invalid device number %d",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }
  if (devNr>(GF46_GetNumDevices()-1)) {
    // No device numbers higher than the available devices.
    sprintf(tmp,"Device %d unavailable",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the (sub)command.
  Tcl_DStringInit(&cmd);
  Tcl_DStringAppend(&cmd,Tcl_GetString(objv[2]),-1);

  // Get the value.
  Tcl_DStringInit(&val);
  Tcl_DStringAppend(&val,Tcl_GetString(objv[3]),-1);

  // Command ladder.
  if (strcmp(Tcl_DStringValue(&cmd),"setAlpha")==0) {
    // setAlpha value
    GF46_SetAlpha(devNr,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setNumeric")==0) {
    // setNumeric value
    GF46_SetNumeric(devNr,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setBrightness")==0) {
    // setBrightness 0..15
    if (Tcl_GetIntFromObj(interp,objv[3], &brightness) != TCL_OK) {
      Tcl_DStringFree(&cmd);
      Tcl_DStringFree(&val);
      return TCL_ERROR;
    }
    GF46_SetBrightness(devNr,brightness);

  } else {
    // Unknown subcommand.
    sprintf(tmp,"Unknown subcommand %s",Tcl_DStringValue(&cmd));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    Tcl_DStringFree(&cmd);
    Tcl_DStringFree(&val);
    return TCL_ERROR;
  }

  Tcl_DStringFree(&cmd);
  Tcl_DStringFree(&val);
  return TCL_OK;
} // gf46Cmd


int gfp8Cmd(ClientData clientData,
            Tcl_Interp *interp,
            int objc,
            Tcl_Obj *CONST objv[]) {
  // Implements all GFP8 (sub)commands.
  // <in> devNr = GFP8 device number, starting with 0.
  // <in> cmd = GFP8 subcommand: setLEDx, setBrightness.
  //            x = 1 for leftmost LED, 8 for rightmost LED.
  // <in> val = 0 or 1.
  // <out> Nothing.

  int devNr;
  char tmp[256];
  Tcl_DString cmd;
  Tcl_DString val;
  int brightness;

  // Check for exactly three parameters.
  if (objc!=4) {
    Tcl_WrongNumArgs(interp,1,objv,"devNr cmd val");
    return TCL_ERROR;
  }

  // Get the device number and check for sanity.
  if (Tcl_GetIntFromObj(interp,objv[1], &devNr) != TCL_OK) {
    return TCL_ERROR;
  }
  if (devNr<0) {
    // No negative device numbers.
    sprintf(tmp,"Invalid device number %d",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }
  if (devNr>(GFP8_GetNumDevices()-1)) {
    // No device numbers higher than the available devices.
    sprintf(tmp,"Device %d unavailable",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the (sub)command.
  Tcl_DStringInit(&cmd);
  Tcl_DStringAppend(&cmd,Tcl_GetString(objv[2]),-1);

  // Get the value.
  Tcl_DStringInit(&val);
  Tcl_DStringAppend(&val,Tcl_GetString(objv[3]),-1);

  // Command ladder.
  if (strcmp(Tcl_DStringValue(&cmd),"setBrightness")==0) {
    // setBrightness 0..15
    if (Tcl_GetIntFromObj(interp,objv[3], &brightness) != TCL_OK) {
      Tcl_DStringFree(&cmd);
      Tcl_DStringFree(&val);
      return TCL_ERROR;
    }
    GFP8_SetBrightness(devNr,brightness);

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED1")==0) {
    // setLED1 value
    setGFP8leds(devNr,0,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED2")==0) {
    // setLED2 value
    setGFP8leds(devNr,1,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED3")==0) {
    // setLED3 value
    setGFP8leds(devNr,2,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED4")==0) {
    // setLED4 value
    setGFP8leds(devNr,3,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED5")==0) {
    // setLED5 value
    setGFP8leds(devNr,4,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED6")==0) {
    // setLED6 value
    setGFP8leds(devNr,5,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED7")==0) {
    // setLED7 value
    setGFP8leds(devNr,6,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED8")==0) {
    // setLED8 value
    setGFP8leds(devNr,7,Tcl_DStringValue(&val));

  } else {
    // Unknown subcommand.
    sprintf(tmp,"Unknown subcommand %s",Tcl_DStringValue(&cmd));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    Tcl_DStringFree(&cmd);
    Tcl_DStringFree(&val);
    return TCL_ERROR;
  }

  Tcl_DStringFree(&cmd);
  Tcl_DStringFree(&val);
  return TCL_OK;
} // gfp8Cmd


int gft8Cmd(ClientData clientData,
            Tcl_Interp *interp,
            int objc,
            Tcl_Obj *CONST objv[]) {
  // Implements all GFT8 (sub)commands.
  // <in> devNr = GFT8 device number, starting with 0.
  // <in> cmd = GFT8 subcommand: setLEDx, setBrightness.
  //            x = 1 for leftmost LED, 8 for rightmost LED.
  // <in> val = 0 or 1.
  // <out> Nothing.

  int devNr;
  char tmp[256];
  Tcl_DString cmd;
  Tcl_DString val;
  int brightness;

  // Check for exactly three parameters.
  if (objc!=4) {
    Tcl_WrongNumArgs(interp,1,objv,"devNr cmd val");
    return TCL_ERROR;
  }

  // Get the device number and check for sanity.
  if (Tcl_GetIntFromObj(interp,objv[1], &devNr) != TCL_OK) {
    return TCL_ERROR;
  }
  if (devNr<0) {
    // No negative device numbers.
    sprintf(tmp,"Invalid device number %d",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }
  if (devNr>(GFT8_GetNumDevices()-1)) {
    // No device numbers higher than the available devices.
    sprintf(tmp,"Device %d unavailable",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the (sub)command.
  Tcl_DStringInit(&cmd);
  Tcl_DStringAppend(&cmd,Tcl_GetString(objv[2]),-1);

  // Get the value.
  Tcl_DStringInit(&val);
  Tcl_DStringAppend(&val,Tcl_GetString(objv[3]),-1);

  // Command ladder.
  if (strcmp(Tcl_DStringValue(&cmd),"setBrightness")==0) {
    // setBrightness 0..15
    if (Tcl_GetIntFromObj(interp,objv[3], &brightness) != TCL_OK) {
      Tcl_DStringFree(&cmd);
      Tcl_DStringFree(&val);
      return TCL_ERROR;
    }
    GFT8_SetBrightness(devNr,brightness);

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED1")==0) {
    // setLED1 value
    setGFT8leds(devNr,0,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED2")==0) {
    // setLED2 value
    setGFT8leds(devNr,1,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED3")==0) {
    // setLED3 value
    setGFT8leds(devNr,2,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED4")==0) {
    // setLED4 value
    setGFT8leds(devNr,3,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED5")==0) {
    // setLED5 value
    setGFT8leds(devNr,4,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED6")==0) {
    // setLED6 value
    setGFT8leds(devNr,5,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED7")==0) {
    // setLED7 value
    setGFT8leds(devNr,6,Tcl_DStringValue(&val));

  } else if (strcmp(Tcl_DStringValue(&cmd),"setLED8")==0) {
    // setLED8 value
    setGFT8leds(devNr,7,Tcl_DStringValue(&val));

  } else {
    // Unknown subcommand.
    sprintf(tmp,"Unknown subcommand %s",Tcl_DStringValue(&cmd));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    Tcl_DStringFree(&cmd);
    Tcl_DStringFree(&val);
    return TCL_ERROR;
  }

  Tcl_DStringFree(&cmd);
  Tcl_DStringFree(&val);
  return TCL_OK;
} // gft8Cmd


int gftq6Cmd(ClientData clientData,
             Tcl_Interp *interp,
             int objc,
             Tcl_Obj *CONST objv[]) {
  // Implements the TQ6 command. Since this is an output-only device,
  // this command does not control the device. Instead it (re)sets
  // the required calibration values.
  // <in> devNr = GF-TQ6 device number, starting with 0.
  // <in> lever = lever to calibrate, levA..levF. Reversers need no
  //              calibration!
  // <in> min = Minimum raw value the lever will ever produce.
  // <in> max = Maximum raw value the lever will ever produce.
  // <in> jit = Jitter the lever generally produces (absolute
  //            deviation in raw data from steady state).
  // <in> steps = Number of steps starting at zero the lever will have
  //              after all calibration has been done.

  int devNr;
  Tcl_DString lever;
  int min,max,jit,steps,levCode;
  char tmp[256];

  // Check for exactly six parameters.
  if (objc!=7) {
    Tcl_WrongNumArgs(interp,1,objv,"devNr lever min max jit steps");
    return TCL_ERROR;
  }

  // Get the device number and check for sanity.
  if (Tcl_GetIntFromObj(interp,objv[1], &devNr) != TCL_OK) {
    return TCL_ERROR;
  }
  if (devNr<0) {
    // No negative device numbers.
    sprintf(tmp,"Invalid device number %d",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }
  if (devNr>(GFTQ6_GetNumDevices()-1)) {
    // No device numbers higher than the available devices.
    sprintf(tmp,"Device %d unavailable",devNr);
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Lever code is parsed last, for code cleanliness reasons.

  // Get the minimum raw lever value.
  if (Tcl_GetIntFromObj(interp,objv[3], &min) != TCL_OK) {
    // Apparently no integer.
    sprintf(tmp,"Illegal minimum raw lever value \"%s\"",
                Tcl_GetString(objv[3]));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the maximum raw lever value.
  if (Tcl_GetIntFromObj(interp,objv[4], &max) != TCL_OK) {
    // Apparently no integer.
    sprintf(tmp,"Illegal maximum raw lever value \"%s\"",
                Tcl_GetString(objv[4]));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // (max-min) should not be zero, to avoid a division by zero
  // condition later on.
  if (max==min) {
    sprintf(tmp,"Minimum and maximum raw lever values are equal",
                Tcl_GetString(objv[4]));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the raw jitter value.
  if (Tcl_GetIntFromObj(interp,objv[5], &jit) != TCL_OK) {
    // Apparently no integer.
    sprintf(tmp,"Illegal jitter value \"%s\"",
                Tcl_GetString(objv[5]));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the number of lever steps.
  if (Tcl_GetIntFromObj(interp,objv[6], &steps) != TCL_OK) {
    // Apparently no integer.
    sprintf(tmp,"Illegal number of steps \"%s\"",
                Tcl_GetString(objv[6]));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    return TCL_ERROR;
  }

  // Get the lever code (string, levA..levF).
  Tcl_DStringInit(&lever);
  Tcl_DStringAppend(&lever,Tcl_GetString(objv[2]),-1);
  // Find out whether a legal lever code has been provided.
  if (strcmp(Tcl_DStringValue(&lever),"levA")==0) {
    levCode = 0;
  } else if (strcmp(Tcl_DStringValue(&lever),"levB")==0) {
    levCode = 1;
  } else if (strcmp(Tcl_DStringValue(&lever),"levC")==0) {
    levCode = 2;
  } else if (strcmp(Tcl_DStringValue(&lever),"levD")==0) {
    levCode = 3;
  } else if (strcmp(Tcl_DStringValue(&lever),"levE")==0) {
    levCode = 4;
  } else if (strcmp(Tcl_DStringValue(&lever),"levF")==0) {
    levCode = 5;
  } else {
    // Unknown lever code.
    sprintf(tmp,"Unknown lever \"%s\" (must be levA..levF)",
                Tcl_DStringValue(&lever));
    Tcl_SetStringObj(Tcl_GetObjResult(interp),tmp,-1);
    Tcl_DStringFree(&lever);
    return TCL_ERROR;
  }

  // Store the calibration for the given lever.
  leverMin[levCode]   = min;
  leverMax[levCode]   = max;
  leverJit[levCode]   = jit;
  leverSteps[levCode] = steps;

  Tcl_DStringFree(&lever);
  return TCL_OK;
} // gftq6Cmd



///// DLL initialisation /////////////////////////////////////////////////

int Gftcl_Init(Tcl_Interp *interp) {
  // Called when the library is loaded into a Tcl interpreter.
  // Initialises the GFdev library.
  // The GFdev library is shut down when the Tcl interpreter is killed.

  // Set up the Tcl stub library, used for version compatibility.
  if (Tcl_InitStubs(interp, "8.2", 0)==NULL) {
    Tcl_SetStringObj(Tcl_GetObjResult(interp),
      "Could not initialise stub library",-1);
    return TCL_ERROR;
	}

  // Initialise the GFdev library.
  if (GFDev_Init(NULL)!=0) {
    Tcl_SetStringObj(Tcl_GetObjResult(interp),
      "Could not initialise GoFlight library",-1);
    return TCL_ERROR;
	}

  // Initialise persistent variables.
  INTERP = interp;
  Tcl_DStringInit(&EVENTPROC);

  // Register the exit handler for the DLL.
  Tcl_CreateExitHandler(done,NULL);

  // Create the new Tcl commands.
  Tcl_CreateObjCommand(interp,
    "gf::version",versionCmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::devices",devicesCmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::notify",notifyCmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::gf166",gf166Cmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::gf45",gf45Cmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::gf46",gf46Cmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::gfp8",gfp8Cmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::gft8",gft8Cmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateObjCommand(interp,
    "gf::gftq6",gftq6Cmd,
    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);

	return TCL_OK;
} // Gftcl_Init


/*
Connect callback.
BOOL MyConnectCallback(GFDEV_HW_TYPE HWType,
                       unsigned short nDevIndex,
                       GFDEV_EVENT_TOKEN event) {
  printf("Connect Callback; device=%d, nr=%d, event=%d\n", HWType,nDevIndex,event); 
  return TRUE;  
}
*/

/*
First attempt at getting the instance handle.
int __stdcall DllMain(void * DLLhandle, unsigned callNr, void * dummy2){
  if (1==callNr) {
printf("DLL init %d %d %d\n",DLLhandle,callNr,dummy2);
    handle = DLLhandle;
    GFDev_RegisterConnectCallback(&MyConnectCallback);
  }
}
*/
